"""
BBOB Benchmark set for Jax - BBOB Benchmark function implemented in JAX
"""

#                                                                       Modules
# =============================================================================

# Standard
from bbob_jax._src.bbob import (
    attractive_sector,
    bent_cigar,
    discuss,
    ellipsoid,
    gallagher_21_peaks,
    gallagher_101_peaks,
    griewank_rosenbrock_f8f2,
    katsuura,
    linear_slope,
    lunacek_bi_rastrigin,
    rastrigin,
    rastrigin_seperable,
    rosenbrock,
    rosenbrock_rotated,
    schaffer_f7_condition_10,
    schaffer_f7_condition_1000,
    schwefel_xsinx,
    sharp_ridge,
    skew_rastrigin_bueche,
    sphere,
    step_ellipsoid,
    sum_of_different_powers,
    weierstrass,
)
from bbob_jax._src.registry import registry, registry_original

#                                                        Authorship and Credits
# =============================================================================
__author__ = "Martin van der Schelling (m.p.vanderschelling@tudelft.nl)"
__credits__ = ["Martin van der Schelling"]
__status__ = "Stable"
#
# =============================================================================


__all__ = [
    "registry",
    "registry_original",
    "bbob",
    "attractive_sector",
    "bent_cigar",
    "discuss",
    "ellipsoid",
    "gallagher_21_peaks",
    "gallagher_101_peaks",
    "griewank_rosenbrock_f8f2",
    "katsuura",
    "linear_slope",
    "lunacek_bi_rastrigin",
    "rastrigin",
    "rastrigin_seperable",
    "rosenbrock",
    "rosenbrock_rotated",
    "schaffer_f7_condition_10",
    "schaffer_f7_condition_1000",
    "schwefel_xsinx",
    "sharp_ridge",
    "skew_rastrigin_bueche",
    "sphere",
    "step_ellipsoid",
    "sum_of_different_powers",
    "weierstrass",
]



#                                                                       Modules
# =============================================================================

# Standard
from collections.abc import Callable
from functools import partial
from typing import Optional

# Third-party
import jax.numpy as jnp
import matplotlib.pyplot as plt
from jax import vmap
from jaxtyping import PRNGKeyArray
from matplotlib.colors import LogNorm

#                                                          Authorship & Credits
# =============================================================================
__author__ = "Martin van der Schelling (M.P.vanderSchelling@tudelft.nl)"
__credits__ = ["Martin van der Schelling"]
__status__ = "Stable"
# =============================================================================


def plot_2d(
    fn: Callable,
    key: PRNGKeyArray,
    bounds: tuple[float, float] = (-5.0, 5.0),
    px: int = 300,
    ax: Optional[plt.Axes] = None,
    log_norm: bool = True,
):
    """
    Plot a 2D heatmap of a function f(x, key) over a given range using imshow.

    Args:
        fn: Callable, function that maps (x, key) -> scalar loss
        key: JAX PRNGKey
        bounds: Tuple (min, max) for x and y axes
        px: Resolution of the mesh grid
        ax: Optional matplotlib Axes
        log_norm: Whether to apply logarithmic normalization to the color scale
    """
    X, Y, Z = create_mesh(fn, key, bounds, px)

    # Create a figure and axis if none provided
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 5))
    else:
        fig = ax.get_figure()

    # Choose normalization
    norm = LogNorm() if log_norm else None

    # Plot with imshow
    _ = ax.imshow(
        Z,
        extent=(*bounds, *bounds),
        origin="lower",
        cmap="viridis",
        norm=norm,
        aspect="auto",
    )

    # Remove ticks for a cleaner look
    ax.set_xticks([])
    ax.set_yticks([])

    return fig, ax


def plot_3d(
    fn: Callable,
    key: PRNGKeyArray,
    bounds: tuple[float, float] = (-5.0, 5.0),
    px: int = 300,
    ax: Optional[plt.Axes] = None,
):
    X, Y, Z = create_mesh(fn, key, bounds, px)

    # Create a figure and axis if none provided
    if ax is None:
        fig = plt.figure(figsize=(7, 5))
        ax = fig.add_subplot(111, projection="3d")
    else:
        fig = ax.get_figure()

    # Plot the surface
    _ = ax.plot_surface(X, Y, Z, cmap="viridis", norm=LogNorm(), zorder=1)

    # Remove ticks for a cleaner look
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])

    return fig, ax


def create_mesh(
    fn: Callable,
    key: PRNGKeyArray,
    bounds: tuple[float, float],
    px: int,
):
    x_vals = jnp.linspace(*bounds, px)
    X, Y = jnp.meshgrid(x_vals, x_vals)

    points = jnp.stack([X.ravel(), Y.ravel()], axis=-1)
    partial_fn = partial(fn, key=key)
    loss_values = vmap(partial_fn)(points)
    Z = loss_values.reshape(X.shape)

    return X, Y, Z



from collections.abc import Callable
from functools import wraps

import jax
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

from bbob_jax._src.utils import fopt, xopt

BBOBFn = Callable[[jax.Array, PRNGKeyArray], jax.Array]

registry_original: dict[str, BBOBFn] = {}

registry: dict[str, BBOBFn] = {}


def register_function(format: str) -> Callable[[BBOBFn], BBOBFn]:
    def decorator(fn: BBOBFn) -> BBOBFn:
        @wraps(fn)
        def wrapper_det(
            x: jax.Array, *args, key: PRNGKeyArray, **kwargs
        ) -> jax.Array:
            x_opt = jnp.zeros(shape=x.shape[-1])
            return fn(x, *args, key=key, x_opt=x_opt, **kwargs)

        @wraps(fn)
        def wrapper_rand(
            x: jax.Array, *args, key: PRNGKeyArray, **kwargs
        ) -> jax.Array:
            x_opt = xopt(key, ndim=x.shape[-1])
            return fn(x, *args, key=key, x_opt=x_opt, **kwargs) + fopt(key)

        # Register both variants
        registry_original[format] = wrapper_det
        registry[format] = wrapper_rand

        return wrapper_rand  # return original function (not wrapped)

    return decorator



from functools import partial

import jax
import jax.numpy as jnp
import jax.random as jr
import pytest

from bbob_jax import registry, registry_original
from bbob_jax.plotting import create_mesh

# Combine both registries into one parameterized source
pytest_registry = [pytest.param(
    name, fn, id=f"registry::{name}") for name, fn in registry.items()]
pytest_registry_original = [
    pytest.param(name, fn, id=f"registry_original::{name}")
    for name, fn in registry_original.items()
]

all_functions = pytest_registry + pytest_registry_original


# Dimensionalities to test
dimensions = [2, 3, 5, 20, 40]


@pytest.mark.parametrize("name,fn", all_functions)
@pytest.mark.parametrize("dim", dimensions)
def test_function_output(name, fn, dim):
    """Test that each registered function runs correctly for given dimensionalities."""
    key = jr.key(0)
    x = jr.uniform(key, shape=(dim,), minval=-5.0, maxval=5.0)

    try:
        y = fn(x, key=key)
    except Exception as e:
        pytest.fail(f"Function {name} raised an exception: {e}")

    assert jnp.isfinite(y), f"Function {name} returned non-finite value: {y}"
    assert jnp.ndim(y) == 0, (
        f"Function {name} did not return a scalar output: {y.shape}"
    )


@pytest.mark.parametrize("name,fn", all_functions)
@pytest.mark.parametrize("dim", dimensions)
def test_function_output_jit(name, fn, dim):
    """Test that each registered function runs correctly for given dimensionalities."""
    key = jr.key(0)
    x = jr.uniform(key, shape=(dim,), minval=-5.0, maxval=5.0)

    try:
        y = jax.jit(fn)(x, key=key)
    except Exception as e:
        pytest.fail(f"Function {name} raised an exception: {e}")

    assert jnp.isfinite(y), f"Function {name} returned non-finite value: {y}"
    assert jnp.ndim(y) == 0, (
        f"Function {name} did not return a scalar output: {y.shape}"
    )


@pytest.mark.parametrize("name,fn", pytest_registry)
@pytest.mark.parametrize("dim", dimensions)
@pytest.mark.parametrize("seed", [1, 2])
def test_function_vmap(name, fn, dim, seed):
    key = jr.key(seed)
    try:
        _, _, Z = create_mesh(fn, key, bounds=(-5.0, 5.0), px=300)
    except Exception as e:
        pytest.fail(f"Function {name} raised an exception during vmap: {e}")

    assert jnp.all(jnp.isfinite(Z)), (
        f"Function {name} returned non-finite values in vmap output."
    )


@pytest.mark.parametrize("name,fn", pytest_registry)
@pytest.mark.parametrize("dim", dimensions)
@pytest.mark.parametrize("seed", [1, 2])
def test_function_grad(name, fn, dim, seed):
    key = jr.key(seed)
    key_x, key_fn = jr.split(key)
    x = jr.uniform(key_x, shape=(300, dim), minval=-5.0, maxval=5.0)
    try:
        grad_fn = jax.grad(partial(fn, key=key_fn))
        grad_value = jax.vmap(grad_fn)(x)
    except Exception as e:
        pytest.fail(
            f"Function {name} raised an exception during jax.grad: {e}")

    assert grad_value.shape == x.shape, (
        f"Function {name} gradient has incorrect shape: "
        f"expected {x.shape}, got {grad_value.shape}"
    )

    assert jnp.all(jnp.isfinite(grad_value)), (
        f"Function {name} returned non-finite gradient values."
    )



#                                                                       Modules
# =============================================================================

# Third-party
import jax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import PRNGKeyArray

# Local
from bbob_jax._src.registry import register_function
from bbob_jax._src.utils import (
    bernoulli_vector,
    lambda_func,
    penalty,
    rotation_matrix,
    tasy_func,
    tosz_func,
)

#                                                          Authorship & Credits
# =============================================================================
__author__ = "Martin van der Schelling (M.P.vanderSchelling@tudelft.nl)"
__credits__ = ["Martin van der Schelling"]
__status__ = "Stable"
# =============================================================================


@register_function("sphere")
def sphere(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array) -> jax.Array:
    z = x - x_opt
    return jnp.sum(z**2)


@register_function("ellipsoid_seperable")
def ellipsoid_seperable(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    i = jnp.arange(1, ndim+1, dtype=x.dtype)
    w = jnp.power(10.0, 6.0 * (i - 1) / (ndim - 1))
    z = tosz_func(x - x_opt)
    return jnp.sum(w * z**2)


@register_function("rastrigin_seperable")
def rastrigin_seperable(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]

    alpha = lambda_func(ndim, alpha=10.0)
    temp = tosz_func(x - x_opt)
    z = jnp.matmul(alpha, tasy_func(temp, beta=0.2))

    return jnp.sum(z**2 - 10.0 * jnp.cos(2.0 * jnp.pi * z) + 10.0 * ndim)


@register_function("skew_rastrigin_bueche")
def skew_rastrigin_bueche(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array):
    ndim = x.shape[-1]
    i = jnp.arange(1, ndim+1, dtype=x.dtype)
    s = jnp.power(10, 0.5 * ((i - 1) / (ndim - 1)))
    odd_indices = jnp.arange(1, ndim+1, 2)

    z = s * tosz_func(x - x_opt)

    # Modify odd indices
    z_odd = jnp.where(
        z[odd_indices] > 0, z[odd_indices] * 10, z[odd_indices]
    )
    z = z.at[odd_indices].set(z_odd)

    # Compute terms
    first_part = 10 * (ndim - jnp.sum(jnp.cos(2.0 * jnp.pi * z)))
    second_part = jnp.sum(z * z)

    y = first_part + second_part + 100 * penalty(x)
    return y


@register_function("linear_slope")
def linear_slope(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    x_opt = 5 * bernoulli_vector(ndim, key)
    i = jnp.arange(1, ndim + 1, dtype=x.dtype)
    s = jnp.sign(x_opt) * jnp.power(10.0, (i - 1) / (ndim - 1))

    cond = x_opt * x < 25.0
    z = jnp.where(cond, x, x_opt)

    result = jnp.sum(5.0 * jnp.abs(s) - s * z)
    return result


@register_function("attractive_sector")
def attractive_sector(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    key1, key2 = jr.split(key)
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    lamb = lambda_func(ndim, alpha=10.0)
    z = Q @ lamb @ R @ (x - x_opt)
    cond = (z * x_opt) > 0.0
    s = jnp.where(cond, 100.0, 1.0)

    term = jnp.sum((s * z) ** 2)

    result = jnp.power(tosz_func(jnp.array([term]))[0], 0.9)

    return result


@register_function("step_ellipsoid")
def step_ellipsoid(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    key1, key2 = jr.split(key)
    ndim = x.shape[-1]
    i = jnp.arange(1, ndim+1, dtype=x.dtype)
    lamb = lambda_func(ndim, alpha=10.0)
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    mult = jnp.power(
        10.0, 2 * ((i - 1) / (ndim - 1))
    )

    # Compute ẑ
    z_hat = lamb @ R @ (x - x_opt)

    # Compute z′ using functional indexing
    z_dash = 0.5 + jnp.where(jnp.abs(z_hat) > 0.5, z_hat, 10 * z_hat)
    z_dash = jnp.floor(z_dash)
    z_dash = jnp.where(jnp.abs(z_hat) > 0.5, z_dash, z_dash / 10.0)

    # Compute z
    z = Q @ z_dash

    # Compute final f
    result = 0.1 * jnp.maximum(
        jnp.abs(z_hat[0]) / 1e4, jnp.sum(mult * z**2)
    )
    return result + penalty(x)


@register_function("rosenbrock")
def rosenbrock(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array) -> jax.Array:
    ndim = x.shape[-1]
    zmax = jnp.maximum(1.0, jnp.sqrt(ndim) / 8.0)
    # Shift and scale
    z = zmax * (x - x_opt) + 1  # shape (..., dim)

    # Create unshifted and shifted arrays along last axis
    unshift = z[..., :-1]  # all except last
    shifted = z[..., 1:]  # all except first

    # Compute the sum
    result = jnp.sum(
        100.0 * jnp.power(unshift**2 - shifted, 2)
        + jnp.power(unshift - 1.0, 2),
        axis=-1,
    )

    return result


@register_function("rosenbrock_rotated")
def rosenbrock_rotated(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key)

    zmax = jnp.maximum(1.0, jnp.sqrt(ndim) / 8.0)

    z = zmax * (x @ R) + 0.5

    # Create unshifted and shifted arrays along last axis
    unshift = z[..., :-1]  # all except last
    shifted = z[..., 1:]  # all except first

    # Compute the sum
    result = jnp.sum(
        100.0 * jnp.power(unshift**2 - shifted, 2)
        + jnp.power(unshift - 1.0, 2),
        axis=-1,
    )

    return result


@register_function("ellipsoid")
def ellipsoid(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array) -> jax.Array:
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key)
    idx = jnp.arange(ndim, dtype=x.dtype)
    z = tosz_func(x @ R)
    weights = 10.0 ** (6.0 * idx / (ndim - 1))
    return jnp.sum(weights * z**2)


@register_function("discuss")
def discuss(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array) -> jax.Array:
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key)
    z = tosz_func(R @ (x - x_opt))
    first = 1e6 * jnp.power(z[..., 0], 2)
    second = jnp.sum(jnp.power(z[..., 1:], 2), axis=-1)
    return first + second


@register_function("bent_cigar")
def bent_cigar(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array) -> jax.Array:
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key)
    z = R @ tasy_func(R @ (x - x_opt), beta=0.5)
    return z[0]**2 + 1e6 * jnp.sum(z[1:]**2)


@register_function("sharp_ridge")
def sharp_ridge(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    key1, key2 = jr.split(key)
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    lamb = lambda_func(ndim, alpha=10.0)
    z = Q @ lamb @ R @ (x - x_opt)
    return z[0]**2 + 100.0 * jnp.sqrt(jnp.sum(z[1:]**2))


@register_function("sum_of_different_powers")
def sum_of_different_powers(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key)
    z = R @ (x - x_opt)
    idx = jnp.arange(1, ndim + 1, dtype=x.dtype)
    return jnp.sum(jnp.abs(z) ** (2 + 4 * (idx - 1) / (ndim - 1)))


@register_function("rastrigin")
def rastrigin(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array) -> jax.Array:
    ndim = x.shape[-1]
    key1, key2 = jr.split(key)
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    lamb = lambda_func(ndim, alpha=10.0)
    z = R @ lamb @ Q @ tasy_func(tosz_func(R @ (x - x_opt)), beta=0.2)

    return jnp.sum(z**2 - 10.0 * jnp.cos(2.0 * jnp.pi * z) + 10.0 * ndim)


@register_function("weierstrass")
def weierstrass(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    key1, key2 = jr.split(key)
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    lamb = lambda_func(ndim, alpha=0.01)
    z = R @ lamb @ Q @ tosz_func(R @ (x - x_opt))

    k = jnp.arange(0, 12, dtype=x.dtype)
    bk = 3.0**k

    f0 = jnp.sum((1 / 2**k) * jnp.cos(2 * jnp.pi * bk * 0.5))

    def inner(z):
        return 1 / 2**k * jnp.cos(2 * jnp.pi * bk * (z + 0.5)) - f0

    y = jax.vmap(inner)(z)
    sum1 = jnp.sum(y)

    first_term = 10.0 * jnp.power((1.0 / ndim) * jnp.sum(sum1), 3)
    pen = (10.0 / ndim) * penalty(x)

    return first_term + pen


@register_function("schaffer_f7_condition_10")
def schaffer_f7_condition_10(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    key1, key2 = jr.split(key)
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    lamb = lambda_func(ndim, alpha=10.0)
    z = lamb @ Q @ tasy_func(R @ (x - x_opt), beta=0.5)

    s = jnp.sqrt(z[:-1] ** 2 + z[1:] ** 2)

    term1 = (1 / (ndim - 1)) * jnp.sum(
        jnp.sqrt(s)
        + jnp.sqrt(s) * jnp.power(jnp.sin(50.0 * jnp.power(s, 0.2)), 2)
    )

    result = jnp.power(term1, 2) + 10 * penalty(x)
    return result


@register_function("schaffer_f7_condition_1000")
def schaffer_f7_condition_1000(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    key1, key2 = jr.split(key)
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    lamb = lambda_func(ndim, alpha=1000.0)
    z = lamb @ Q @ tasy_func(R @ (x - x_opt), beta=0.5)

    s = jnp.sqrt(z[:-1] ** 2 + z[1:] ** 2)

    term1 = (1 / (ndim - 1)) * jnp.sum(
        jnp.sqrt(s)
        + jnp.sqrt(s) * jnp.power(jnp.sin(50.0 * jnp.power(s, 0.2)), 2)
    )

    result = jnp.power(term1, 2) + 10 * penalty(x)
    return result


@register_function("griewank_rosenbrock_f8f2")
def griewank_rosenbrock_f8f2(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key)
    z = jnp.maximum(1.0, jnp.sqrt(ndim) / 8.0) * (R @ x) + 0.5
    s = 100 * (z[:-1] ** 2 - z[1:]) ** 2 + (z[:-1] - 1) ** 2

    return (10 / (ndim - 1)) * jnp.sum((s / 4000) - jnp.cos(s)) + 10.0


@register_function("schwefel_xsinx")
def schwefel_xsinx(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    ndim = x.shape[-1]
    ones = bernoulli_vector(ndim, key)
    lamb = lambda_func(ndim, alpha=10.0)
    x_hat = 2.0 * ones * x
    x_opt = 4.2096874633 / 2 * ones

    z_hat = x_hat.at[..., 1:].add(
        0.25 * (x_hat[..., :-1] - 2.0 * jnp.abs(x_opt[..., :-1]))
    )

    z = 100.0 * (lamb @ (z_hat - 2.0 * jnp.abs(x_opt)) + 2.0 * jnp.abs(x_opt))

    f = (
        -1.0
        / (100.0 * ndim)
        * jnp.sum(z * jnp.sin(jnp.sqrt(jnp.abs(z))), axis=-1)
    )

    # Penalization
    pen = 100.0 * penalty(z / 100.0)

    return f + 4.189828872724339 + pen


@register_function("gallagher_101_peaks")
def gallagher_101_peaks(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    key1, key2, key3 = jr.split(key, 3)

    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key1)
    i = jnp.arange(1, 102, dtype=jnp.float32)
    j = jnp.arange(0, 100, dtype=jnp.float32)

    w = 1.1 + 8.0 * ((i - 2) / 99)
    w = w.at[0].set(10.0)

    a = jnp.power(1000, 2.0 * (j / 99.0))
    alpha = jr.permutation(key2, a)
    alpha = jnp.concatenate([jnp.array([1000.0]), alpha])

    y = jr.uniform(key3, shape=(*i.shape, ndim), minval=-5.0, maxval=5.0)
    x_opt = jr.uniform(key, shape=(ndim,), minval=-4.0, maxval=4.0)
    y = y.at[0].set(x_opt)

    C = jax.vmap(lambda_func, in_axes=(None, 0))(ndim, alpha)

    C /= jnp.power(alpha, 0.25)[:, None, None]

    diff = x[None, :] - y

    def apply_C(C_i, d_i, w_i):
        val = -(1.0 / (2.0 * ndim)) * d_i.T @ (R.T @ C_i @ R) @ d_i
        return w_i * jnp.exp(val)

    inside_max = jax.vmap(apply_C)(C, diff, w)

    f = 10.0 - jnp.max(inside_max, axis=0)

    f_tosz = tosz_func(jnp.array([f]))[0]

    result = jnp.power(f_tosz, 2) + penalty(x)

    return result


@register_function("gallagher_21_peaks")
def gallagher_21_peaks(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array
) -> jax.Array:
    key1, key2, key3 = jr.split(key, 3)

    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key1)
    i = jnp.arange(1, 22, dtype=jnp.float32)
    j = jnp.arange(0, 20, dtype=jnp.float32)

    w = 1.1 + 8.0 * ((i - 2) / 19)
    w = w.at[0].set(10.0)

    a = jnp.power(1000, 2.0 * (j / 19.0))
    alpha = jr.permutation(key2, a)
    alpha = jnp.concatenate([jnp.array([1000.0**2]), alpha])

    y = jr.uniform(key3, shape=(*i.shape, ndim), minval=-4.9, maxval=4.9)
    x_opt = jr.uniform(key, shape=(ndim,), minval=-3.92, maxval=3.92)
    y = y.at[0].set(x_opt)

    C = jax.vmap(lambda_func, in_axes=(None, 0))(ndim, alpha)

    C /= jnp.power(alpha, 0.25)[:, None, None]

    diff = x[None, :] - y

    def apply_C(C_i, d_i, w_i):
        val = -(1.0 / (2.0 * ndim)) * d_i.T @ (R.T @ C_i @ R) @ d_i
        return w_i * jnp.exp(val)

    inside_max = jax.vmap(apply_C)(C, diff, w)

    f = 10.0 - jnp.max(inside_max, axis=0)

    f_tosz = tosz_func(jnp.array([f]))[0]

    result = jnp.power(f_tosz, 2) + penalty(x)

    return result


@register_function("katsuura")
def katsuura(x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array) -> jax.Array:
    key1, key2 = jr.split(key)
    ndim = x.shape[-1]
    R = rotation_matrix(ndim, key1)
    Q = rotation_matrix(ndim, key2)
    lamb = lambda_func(ndim, alpha=100.0)
    z = Q @ lamb @ R @ (x - x_opt)

    J = 2.0 ** jnp.arange(1, 33, dtype=jnp.float32)  # (32,)
    # jsum term: shape (32, dim)
    z_expanded = z[None, :]  # (1, dim)
    J_expanded = J[:, None]  # (32, 1)
    jsum = (
        jnp.abs(J_expanded * z_expanded - jnp.round(J_expanded * z_expanded))
        / J_expanded
    )  # (32, dim)

    # Sum over j (the 32 terms)
    sum_j = jnp.sum(jsum, axis=0)  # (dim,)

    # Multiply by (1..dim) and add 1
    bracket = 1.0 + jnp.arange(1, ndim + 1, dtype=x.dtype) * sum_j  # (dim,)
    prod = jnp.prod(bracket)

    # Final scaling and power
    prod = jnp.power(prod, 10.0 / ndim**1.2)
    prod = prod * (10.0 / ndim**2.0) - (10.0 / ndim**2.0)

    return prod + penalty(x)


@register_function("lunacek_bi_rastrigin")
def lunacek_bi_rastrigin(
    x: jax.Array, key: PRNGKeyArray, x_opt: jax.Array


) -> jax.Array:
    ndim = x.shape[-1]
    key1, key2, key3 = jr.split(key, 3)
    mu0 = 2.5
    d = 1.0
    s = 1.0 - 1.0 / (2.0 * jnp.sqrt(ndim + 20.0) - 8.2)

    x_opt = (mu0 / 2.0) * bernoulli_vector(ndim, key1)
    x_hat = 2 * jnp.sign(x_opt) * x

    R = rotation_matrix(ndim, key2)
    Q = rotation_matrix(ndim, key3)
    lamb = lambda_func(ndim, alpha=100.0)
    z = Q @ lamb @ R @ (x_hat - mu0 * jnp.ones_like(x))

    mu1 = -jnp.sqrt((mu0**2 - d) / s)

    term1 = jnp.minimum(
        jnp.sum(jnp.power(x_hat - mu0, 2)),
        d * ndim + s * jnp.sum(jnp.power(x_hat - mu1, 2)),
    )

    term2 = 10.0 * (ndim - jnp.sum(jnp.cos(2.0 * jnp.pi * z)))

    return term1 + term2 + 1e4 * penalty(x)



import jax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import PRNGKeyArray


def fopt(key: PRNGKeyArray) -> jax.Array:
    """Generate a random optimal function value f_opt."""
    return jnp.round(jnp.clip(100.0 * jr.cauchy(key), -1000.0, 1000.0), 2)


def xopt(key: PRNGKeyArray, ndim: int) -> jax.Array:
    """Generate a random optimal solution x_opt within [-4, 4]^ndim."""
    return jr.uniform(key, shape=(ndim,), minval=-5.0, maxval=5.0)


def tosz_func(x):
    c1, c2 = 10.0, 7.9
    eps = 1e-12  # avoid log(0)

    x = jnp.asarray(x)
    abs_x = jnp.maximum(jnp.abs(x), eps)
    x_sign = jnp.sign(x)
    x_star = jnp.log(abs_x)
    transformed = x_sign * jnp.exp(
        x_star + 0.049 * (jnp.sin(c1 * x_star) + jnp.sin(c2 * x_star))
    )

    # same “special treatment” as original, but now applied elementwise
    mask = (x == x[0]) | (x == x[-1])
    return jnp.where(mask, transformed, x)


def tasy_func(x: jax.Array, beta=0.5) -> jax.Array:
    ndim = x.shape[-1]
    idx = jnp.arange(0, ndim)
    up = 1 + beta * ((idx - 1) / (ndim - 1)) * jnp.sqrt(jnp.abs(x))
    x_temp = jnp.abs(x) ** up
    return jnp.where(x > 0, x_temp, x)


def lambda_func(size: int, alpha: float = 10.0) -> jax.Array:
    idx = jnp.arange(size, dtype=jnp.float32)
    diagonal = alpha ** (idx / (2 * (size - 1)))
    return jnp.diag(diagonal)


def rotation_matrix(dim: int, key: jax.Array) -> jax.Array:
    """Generate a random orthogonal rotation matrix."""
    R = jr.normal(key, shape=(dim, dim))
    Q, R_ = jnp.linalg.qr(R)
    # Ensure a right-handed coordinate system (determinant = +1)
    d = jnp.sign(jnp.linalg.det(Q))
    Q = Q * d
    return Q


def penalty(x: jax.Array) -> jax.Array:
    return jnp.sum(jnp.power(jnp.maximum(jnp.abs(x) - 5.0, 0.0), 2), axis=-1)


def bernoulli_vector(dim: int, key: jax.Array) -> jax.Array:
    """Generate a random Bernoulli matrix with entries -1 or 1."""
    return jr.bernoulli(key, p=0.5, shape=(dim,)).astype(jnp.float32) * 2 - 1
